/****************************************************************************\
*                                                                            *
*  FISH-HC                                                                   *
*                                                                            *
*  High color Fastgraph fish tank. Requires Win32.                           *
*                                                                            *
\****************************************************************************/

#include <fgwin.h>
#include <io.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>

#define WIDTH  640
#define HEIGHT 480
#define DEPTH  16

#define RIGHT 0
#define LEFT  1

#define UP  -1
#define DOWN 1

/* generate a random integer between 2 numbers */
#define IRAND(min,max) ((rand()%((max)-(min)+1))+(min))

/* location of each fish frame as stored in a PCX file */
typedef struct _frame
{
   int x1; int x2; int y1; int y2;
} FRAME;

/* type FISH defines each species of fish */
typedef struct _fish
{
   char  *fname;
   int   direction;
   int   nframes;
   int   frame_delay;
   FRAME frame[4];
   int   width[4];
   int   height[4];
   short *bitmap[4];
} FISH;

FISH fish[6] =
{{ "BLUEDAMS.PCX", LEFT, 4, 4,
    79,232,218,281,
   270,423,219,281,
   457,610,219,281,
    73,226,334,396,
   0,0,0,0,
   0,0,0,0,
   NULL,NULL,NULL,NULL
},{
   "BLUETANG.PCX", RIGHT, 4, 4,
   492,635,200,275,
   325,468,200,275,
   165,308,200,275,
     1,144,199,274,
   0,0,0,0,
   0,0,0,0,
   NULL,NULL,NULL,NULL
},{
   "BUTTRFLY.PCX", RIGHT, 4, 4,
   244,397,191,286,
    41,194,191,286,
    41,194,337,433,
   244,397,338,433,
   0,0,0,0,
   0,0,0,0,
   NULL,NULL,NULL,NULL
},{
   "GUDGEON.PCX", RIGHT, 4, 4,
    23,203,198,280,
   231,411,195,280,
   438,618,195,280,
    23,203,340,425,
   0,0,0,0,
   0,0,0,0,
   NULL,NULL,NULL,NULL
},{
   "KILLIFSH.PCX", RIGHT, 4,  4,
   133,285,153,205,
   350,502,153,205,
   136,288,273,324,
   350,502,272,324,
   0,0,0,0,
   0,0,0,0,
   NULL,NULL,NULL,NULL
},{
   "SEAHORSE.PCX", LEFT, 3, 4,
   388,435,205,293,
   295,342,205,293,
   199,246,205,293,
   0,0,0,0,
   0,0,0,0,
   0,0,0,0,
   NULL,NULL,NULL,NULL
}};

/* forward declaration */
struct _fishlist;
typedef struct _fishlist FISHLIST;

/* Each fish is stored as a sprite in a linked list */
typedef struct _fishlist
{
   FISHLIST *next;
   int fishnum;
   int x;
   int y;
   int xmin;
   int xmax;
   int xinc;
   int ymin;
   int ymax;
   int yinc;
   int frame;
   int x_direction;
   int y_direction;
   int f_direction;
   int frame_counter;
} FISHLIST;

FISHLIST *head;
FISHLIST *tail;

int background; /* background virtual buffer */
int work;       /* workspace virtual buffer */
int wheel;      /* water wheel virtual buffer */

/* buffers required for water wheel flic file */
char wheel_context[20];
char *image_buf;

int counter = 0;
int bubble_frame = 0;

/* function declarations */
void animate_fish(void);
void get_fish(int);
void init_fish(void);

long WINAPI _export WindowProc(HWND,UINT,UINT,LONG);

int PASCAL WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    LPSTR lpszCmdParam, int nCmdShow)
{
   static char szAppName[] = "FGfish";
   HWND        window;
   MSG         msg;
   WNDCLASS    wndclass;

   /* set up the window class */
   if (!hPrevInstance)
   {
      wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
      wndclass.lpfnWndProc   = WindowProc;
      wndclass.cbClsExtra    = 0;
      wndclass.cbWndExtra    = 0;
      wndclass.hInstance     = hInstance;
      wndclass.hIcon         = LoadIcon(NULL,IDI_APPLICATION);
      wndclass.hCursor       = LoadCursor(NULL,IDC_ARROW);
      wndclass.hbrBackground = NULL;
      wndclass.lpszMenuName  = NULL;
      wndclass.lpszClassName = szAppName;
      RegisterClass (&wndclass);
   }

   /* create and initially display the window */
   window = CreateWindow(szAppName,     // window class name
      "Fastgraph High Color Fish Tank", // window caption
      WS_POPUP,                // window style
      0,                       // initial x position
      0,                       // initial y position
      WIDTH,                   // initial x size
      HEIGHT,                  // initial y size
      NULL,                    // parent window handle
      NULL,                    // window menu handle
      hInstance,               // program instance handle
      NULL);                   // creation parameters

   ShowWindow(window,nCmdShow);
   UpdateWindow(window);

   /* The message loop processes entries placed in the message queue.
      When no message is ready, call animate_fish() to perform one frame
      of animation. */

   while (TRUE)
   {
      if (PeekMessage(&msg,NULL,0,0,PM_REMOVE))
      {
         if (msg.message == WM_QUIT)
            break;
         else
         {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
         }
      }
      else
         animate_fish();
   }

   return msg.wParam;
}

/****************************************************************************\
*                                                                            *
*  WindowProc                                                                *
*                                                                            *
*  Window procedure to handle messages sent to the window.                   *
*                                                                            *
\****************************************************************************/

HDC      hdc;
HPALETTE hpal;

long WINAPI _export WindowProc(HWND window, UINT message,
                               UINT wParam, LONG lParam)
{
   PAINTSTRUCT ps;
   int  button;
   int  buffsize;
   char *buffer;
   char header[10];
   FILE *stream;
   register int i;

   switch (message)
   {
      case WM_CREATE:
         hdc = GetDC(window);
         fg_setdc(hdc);
         hpal = fg_defpal();
         fg_realize(hpal);
         ShowWindow(window,SW_SHOWNORMAL);

         /* set 640x480 screen resolution */
         if (fg_modetest(WIDTH,HEIGHT,fg_colors()) != 0)
         {
            MessageBox(window,"Cannot set desktop resolution.","Error",MB_OK|MB_ICONSTOP);
            DestroyWindow(window);
            return 0;
         }
         fg_modeset(WIDTH,HEIGHT,fg_colors(),1);

         /* initialize virtual buffers */
         fg_vbinit();
         fg_vbdepth(DEPTH);
         wheel = fg_vballoc(186,153);
         work = fg_vballoc(WIDTH,HEIGHT);
         background = fg_vballoc(WIDTH,HEIGHT);

         /* load background image into the background virtual buffer */
         fg_jpeghead("AQUARIUM.JPG",header);
         buffsize = fg_jpegmem(header);
         buffer = (char *)malloc(buffsize);
         fg_jpegbuf(buffer,buffsize);
         fg_vbopen(background);
         fg_showjpeg("AQUARIUM.JPG",0);
         free(buffer);

         /* get all the fish and put them into a linked list */
         fg_vbopen(work);
         for (i = 0; i < 6; i++)
            get_fish(i);
         init_fish();

         /* load the water wheel flic file into the fg_imagebuf() buffer */
         fg_flicopen("WHEEL3.FLC",wheel_context);
         fg_flicdone(wheel_context);
         stream = fopen("WHEEL3.FLC","rb");
         buffsize = filelength(fileno(stream));
         image_buf = (char *)malloc(buffsize);
         fread(image_buf,sizeof(char),buffsize,stream);
         fclose(stream);
         fg_imagebuf(image_buf,buffsize);

         /* load virtual palette with the water wheel flic file palette */
         fg_vbopen(wheel);
         fg_flicplay(wheel_context,1,5);
         fg_vbclose();

         /* things won't look too good with a 256-color display driver */
         if (fg_colors() <= 8)
         {
            MessageBeep(0);
            button = MessageBox(window,
               "A high color or true color display is recommended. Continue?",
               "FISH",MB_ICONQUESTION|MB_YESNO|MB_DEFBUTTON2);
            if (button == IDNO)
               PostMessage(window,WM_CLOSE,0,0);
         }

         fg_vbopen(background);
         return 0;

      case WM_KEYDOWN:
         switch(wParam)
         {
            case VK_ESCAPE:
            case VK_F12:
               DestroyWindow(window);
               break;
         }
         return 0;

      case WM_PAINT:
         BeginPaint(window,&ps);
         fg_vbpaste(0,WIDTH-1,0,HEIGHT-1,0,HEIGHT-1);
         EndPaint(window,&ps);
         return 0;

      case WM_SETFOCUS:
         fg_realize(hpal);
         InvalidateRect(window,NULL,TRUE);
         return 0;

      case WM_DESTROY:
         fg_vbclose();
         fg_vbfree(background);
         fg_vbfree(wheel);
         fg_vbfree(work);
         fg_vbfin();
         DeleteObject(hpal);
         ReleaseDC(window,hdc);
         fg_modeset(0,0,0,0);
         PostQuitMessage(0);
         return 0;
   }
   return DefWindowProc(window,message,wParam,lParam);
}

/****************************************************************************/
void animate_fish(void)
{
   FISHLIST *node;
   register int i;
   int rows, y;
   int r, g, b;

   short bubble[14*14];
   static char bubble_mask[14*14] = {
      0,0,0,0,0,1,1,1,1,0,0,0,0,0,
      0,0,0,1,1,1,1,1,1,1,1,0,0,0,
      0,0,1,1,1,1,1,1,1,1,1,1,0,0,
      0,0,1,1,1,1,1,1,1,1,1,1,0,0,
      0,1,1,1,1,1,1,1,1,1,1,1,1,0,
      0,1,1,1,1,1,1,1,1,1,1,1,1,0,
      1,1,1,1,1,1,1,1,1,1,1,1,1,1,
      1,1,1,1,1,1,1,1,1,1,1,1,1,1,
      0,1,1,1,1,1,1,1,1,1,1,1,1,0,
      0,1,1,1,1,1,1,1,1,1,1,1,1,0,
      0,0,1,1,1,1,1,1,1,1,1,1,0,0,
      0,0,1,1,1,1,1,1,1,1,1,1,0,0,
      0,0,0,1,1,1,1,1,1,1,1,0,0,0,
      0,0,0,0,0,1,1,1,1,0,0,0,0,0};

   /* copy background to workspace */
   fg_vbcopy(0,WIDTH-1,0,HEIGHT-1,0,HEIGHT-1,background,work);

   /* display next frame of water wheel flic file */
   fg_vbopen(wheel);
   if (fg_flicplay(wheel_context,1,5) == 0)
   {
      fg_flicskip(wheel_context,-1);
      fg_flicskip(wheel_context,1);
      fg_flicplay(wheel_context,1,5);
      bubble_frame = 0;
   }
   fg_vbcopy(0,185,0,152,WIDTH-186,HEIGHT-1,wheel,work);

   /* create and display a translucent bubble moving up */
   fg_vbopen(work);
   bubble_frame++;
   y = (33 - bubble_frame) * 11;
   fg_move(IRAND(457,459),y);
   rows = min(y,14);
   if (rows > 0)
   {
      fg_getdcb(bubble,14,rows);
      for (i = 0; i < 14*14; i++)
      {
         if (bubble_mask[i])
         {
            fg_unmaprgb(bubble[i],&r,&g,&b);
            bubble[i] = fg_maprgb(r+30,g+30,b+30);
         }
      }
      fg_clipdcb(bubble,14,rows);
   }

   /* display all the fish in the linked list */
   for (node = head; node != (FISHLIST *)NULL; node = node->next)
   {
      /* blit the fish, may reverse while blitting */
      fg_move(node->x,node->y);
      if (node->x_direction == fish[node->fishnum].direction)
      {
         fg_clipdcb(fish[node->fishnum].bitmap[node->frame],
            fish[node->fishnum].width[node->frame],
            fish[node->fishnum].height[node->frame]);
      }
      else
      {
         fg_flipdcb(fish[node->fishnum].bitmap[node->frame],
            fish[node->fishnum].width[node->frame],
            fish[node->fishnum].height[node->frame]);
      }

      /* increment the horizontal position */
      if (node->x_direction == RIGHT)
      {
         node->x += node->xinc;
         if (node->x > node->xmax)
            node->x_direction = LEFT;
      }
      else
      {
         node->x -= node->xinc;
         if (node->x < node->xmin)
            node->x_direction = RIGHT;
      }

      /* adjust vertical position */
      node->y += node->y_direction;

      /* keep vertical motion within tolerance area */
      if (node->y >= node->ymax)
         node->y_direction = UP;
      else if (node->y <= node->ymin)
         node->y_direction = DOWN;

      /* add a random element to speed and vertical motion */
      counter++;
      if (counter % node->y == 0)
      {
         node->y_direction = IRAND(-1,5);
         if (node->y_direction > 1)
            node->y_direction = 0;
         node->xinc = IRAND(1,4);
         node->ymin = IRAND(60,400);
         node->ymax = IRAND(node->ymin,HEIGHT-1);
      }

      /* increment fish frame (swish tail slowly) */
      node->frame_counter++;
      if (node->frame_counter >= fish[node->fishnum].frame_delay)
      {
         if (node->f_direction == RIGHT)
         {
            node->frame++;
            if (node->frame >= fish[node->fishnum].nframes-1)
               node->f_direction = LEFT;
         }
         else
         {
            node->frame--;
            if (node->frame <= 1)
               node->f_direction = RIGHT;
         }
         node->frame_counter = 0;
      }
   }

   /* display the frame we just constructed */
   fg_vbpaste(0,WIDTH-1,0,HEIGHT-1,0,HEIGHT-1);
}

/****************************************************************************/
void get_fish(int fishnum)
{
   register int i;

   /* display the fish PCX file the workspace virtual buffer */
   fg_showpcx(fish[fishnum].fname,0);

   /* grab the bitmaps for all the frames */
   for (i = 0; i < fish[fishnum].nframes; i++)
   {
      fish[fishnum].width[i] =
         fish[fishnum].frame[i].x2 - fish[fishnum].frame[i].x1 + 1;
      fish[fishnum].height[i] =
         fish[fishnum].frame[i].y2 - fish[fishnum].frame[i].y1 + 1;

      fish[fishnum].bitmap[i] =
         (short *)malloc(fg_imagesiz(fish[fishnum].width[i],fish[fishnum].height[i]));

      fg_move(fish[fishnum].frame[i].x1,fish[fishnum].frame[i].y2);
      fg_getdcb(fish[fishnum].bitmap[i],
         fish[fishnum].width[i],fish[fishnum].height[i]);
   }
}

/****************************************************************************/
void init_fish(void)
{
   FISHLIST *node;
   int nodes;
   register int i;

   /* choose a random number of fish */
   srand(fg_getclock());
   nodes = IRAND(6,12);

   /* build a linked list */
   for (i = 0; i < nodes; i++)
   {
      node = (FISHLIST *) malloc(sizeof(FISHLIST));
      node->x_direction = IRAND(RIGHT,LEFT);
      node->y_direction = IRAND(UP,DOWN);
      node->f_direction = IRAND(0,1);
      node->fishnum     = IRAND(0,5);
      node->next        = (FISHLIST *)NULL;
      node->xmin        = IRAND(-700,-200);
      node->xmax        = IRAND(700,1200);
      if (node->x_direction == RIGHT)
         node->x        = node->xmin;
      else
         node->x        = node->xmax;
      node->xinc        = IRAND(1,4);
      node->ymin        = IRAND(60,400);
      node->ymax        = IRAND(node->ymin,479);
      node->y           = IRAND(node->ymin,node->ymax);
      node->yinc        = IRAND(1,4);
      node->frame       = IRAND(1,2);
      node->frame_counter   = IRAND(0,2);
      if (head == (FISHLIST *)NULL)
      {
         head = node;
         tail = head;
      }
      else
      {
         tail->next = node;
         tail = node;
      }
   }
}
